/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/geometry/Utils.cpp

  \brief Utility functions for the Geometry Module.
*/

// TerraLib
#include "../common/STLUtils.h"
#include "../core/translator/Translator.h"
#include "Envelope.h"
#include "Exception.h"
#include "Geometry.h"
#include "GeometryCollection.h"
#include "GEOSGeometryFactory.h"
#include "GEOSReader.h"
#include "GEOSWriter.h"
#include "Line.h"
#include "LinearRing.h"
#include "LineString.h"
#include "MultiLineString.h"
#include "MultiPolygon.h"
#include "Point.h"
#include "Polygon.h"
#include "Utils.h"

// STL
#include <cmath>

#ifdef TERRALIB_GEOS_ENABLED
// GEOS
#include <geos/algorithm/CentroidArea.h>
#include <geos/algorithm/CGAlgorithms.h>
#include <geos/geom/Coordinate.h>
#include <geos/geom/Geometry.h>
#include <geos/geom/LineSegment.h>
#include <geos/geom/Point.h>
#include <geos/geom/Polygon.h>
#include <geos/operation/polygonize/Polygonizer.h>
#include <geos/operation/union/CascadedPolygonUnion.h>
#include <geos/operation/union/UnaryUnionOp.h>
#include <geos/operation/distance/DistanceOp.h>
#include <geos/operation/overlay/snap/GeometrySnapper.h>
#include <geos/operation/valid/IsValidOp.h>
#include <geos/util/GEOSException.h>
#endif

const double tolerance = 0.000000001;

std::unique_ptr<te::gm::Geometry> te::gm::GetGeometryUnion(const std::vector<gm::Geometry*> &geomVec)
{
  std::unique_ptr<te::gm::Geometry> geometry;

  if (geomVec.size() == 1)
  {
    te::dt::AbstractData* abs = geomVec[0]->clone();
    geometry.reset(static_cast<te::gm::Geometry*>(abs));
  }
  else if (geomVec.size() > 1)
  {
    //First clone the vector
    std::vector<te::gm::Geometry*> geomVecCopy;
    for (std::size_t i = 0; i < geomVec.size(); ++i)
    {
      geomVecCopy.push_back(static_cast<te::gm::Geometry*>(geomVec[i]->clone()));
    }

    for (std::size_t i = 0; i < geomVecCopy.size(); ++i)
    {
      std::unique_ptr<te::gm::Geometry> currentGeom(static_cast<te::gm::Geometry*>(geomVecCopy[i]->clone()));

      for(std::size_t j = 0; j < geomVecCopy.size(); ++j)
      {
        if(i != j)
        {
          currentGeom.reset(te::gm::PrepareGeometriesToIntersection(currentGeom.get(), geomVecCopy[j]));
        }
      }

      //Must always keep the most updated version of the geometry. It can be reused by the algorithm.
      delete geomVecCopy[i];
      geomVecCopy[i] = currentGeom.release();
    }

    std::unique_ptr<te::gm::GeometryCollection> gc(new te::gm::GeometryCollection(0, te::gm::GeometryCollectionType, geomVec[0]->getSRID()));

    for (std::size_t i = 0; i < geomVecCopy.size(); ++i)
      gc->add(geomVecCopy[i]);

    bool success = false;

    try
    {
      geometry.reset(te::gm::UnaryUnion(gc.get()));
      success = true;
    }
    catch(...)
    {
    }

    if(success == false)
    {
      try
      {
        geometry.reset(te::gm::UnaryUnion(gc->buffer(0)));
      }
      catch(geos::util::GEOSException& e)
      {
        throw Exception(e.what());
      }
    }
  }

  return geometry;
}

te::gm::Geometry* te::gm::GetGeomFromEnvelope(const Envelope* const e, int srid)
{
// create an outer ring with the same envelope as our envelope
  LinearRing* r = new LinearRing(5, LineStringType, srid, new Envelope(*e));

  r->setPoint(0, e->m_llx, e->m_lly);
  r->setPoint(1, e->m_urx, e->m_lly);
  r->setPoint(2, e->m_urx, e->m_ury);
  r->setPoint(3, e->m_llx, e->m_ury);
  r->setPoint(4, e->m_llx, e->m_lly);

// create the polygon
  Polygon* p = new Polygon(1, PolygonType, srid, new Envelope(*e));
  p->setRingN(0, r);

  return p;
}

bool te::gm::SatisfySpatialRelation(const Geometry* g1,
                                    const Geometry* g2, 
                                    SpatialRelation relation)
{
  switch(relation)
  {
    case CONTAINS:
      return g1->contains(g2);

    case COVEREDBY:
      return g1->coveredBy(g2);

    case COVERS:
      return g1->covers(g2);

    case CROSSES:
      return g1->crosses(g2);

    case DISJOINT:
      return g1->disjoint(g2);

    case EQUALS:
      return g1->equals(g2);

    case INTERSECTS:
      return g1->intersects(g2);

    case OVERLAPS:
      return g1->overlaps(g2);

    case TOUCHES:
      return g1->touches(g2);

    case WITHIN:
      return g1->within(g2);

    default:
      throw Exception(TE_TR("Invalid spatial relation!"));
  }
}

te::gm::Envelope te::gm::AdjustToCut(const Envelope & env, double bWidth, double bHeight)
{
  double auxD;
  int auxI;

  int magicX; 
  auxD = env.m_llx/bWidth;
  auxI = (int)(env.m_llx/bWidth);
  if (env.m_llx < 0 && (auxD - auxI) != 0)
    magicX = (int) (env.m_llx/bWidth - 1);
  else
    magicX = auxI;

  int magicY;
  auxD = env.m_lly/bHeight;
  auxI = (int)(env.m_lly/bHeight);
  if (env.m_lly < 0 && (auxD - auxI) != 0)
    magicY  = (int)(env.m_lly/bHeight - 1);
  else
    magicY  = auxI;

  double xi = magicX*bWidth;
  double yi = magicY*bHeight;

  int magicX2;
  auxD = env.m_urx/bWidth;
  auxI = (int)(env.m_urx/bWidth);
  if ((env.m_urx < 0) || (auxD - auxI) == 0)
    magicX2 = (int) (env.m_urx/bWidth);
  else
    magicX2 = (int) (env.m_urx/bWidth + 1);

  int magicY2;
  auxD = env.m_ury/bHeight;
  auxI = (int)(env.m_ury/bHeight);
  if ((env.m_ury < 0) || (auxD - auxI) == 0)
    magicY2 = (int) (env.m_ury/bHeight);
  else
    magicY2 = (int) (env.m_ury/bHeight + 1);

  double xf = (magicX2)*bWidth;
  double yf = (magicY2)*bHeight;

  return te::gm::Envelope(xi,yi,xf,yf);
}

template<class T1, class T2> bool te::gm::Intersects(const T1& o1, const T2& o2)
{
  return o1->intersects(o2);
}

template<> bool te::gm::Intersects(const te::gm::Point& point, const te::gm::Envelope& e)
{
  // point to the right of envelope
  if(point.getX() > e.m_urx)
    return false;

  // point to the left of envelope
  if(e.m_llx > point.getX())
    return false;

  // point is above envelope
  if(point.getY() > e.m_ury)
    return false;

  // point is below envelope
  if(e.m_lly > point.getY())
    return false;

  return true;
}

te::gm::Coord2D* te::gm::locateAlong(const LineString* line, double initial, double final, double target)
{
  double tTof; // Distance of target to fist point

  std::map<int, double> pointLenghtFromFirst;

  double fullLineLenght = 0;
  
  pointLenghtFromFirst[0] = 0;
  for(std::size_t i = 1; i < line->getNPoints(); i++)
  {
    std::unique_ptr<Point> pA = line->getPointN(i);
    std::unique_ptr<Point> pB = line->getPointN(i - 1);

    fullLineLenght += pA->distance(pB.get());
    pointLenghtFromFirst[i] = fullLineLenght;
  }

  double diference = final-initial;
  
  tTof = ((target-initial)*fullLineLenght)/diference;

  int onTheFly = -1;

  int pointBeforeTarget;
  for(std::size_t i = 0; i < pointLenghtFromFirst.size(); i++)
  {
    if(pointLenghtFromFirst[i] == tTof)
    {
      onTheFly = i;
      break;
    }
    if(tTof >= pointLenghtFromFirst[i] && tTof < pointLenghtFromFirst[i+1])
    {
      pointBeforeTarget = i;
      break;
    }
  }

  te::gm::Coord2D* targetCoord = new te::gm::Coord2D();

  if(onTheFly >= 0)
  {
    targetCoord->x = line->getPointN(onTheFly)->getX();
    targetCoord->y = line->getPointN(onTheFly)->getY();

    return targetCoord;
  }

  std::unique_ptr<Point> p1 = line->getPointN(pointBeforeTarget);
  std::unique_ptr<Point> p2 = line->getPointN(pointBeforeTarget + 1);

  // Partial distance: target to p1
  double pd = tTof - pointLenghtFromFirst[pointBeforeTarget];
  
  // Total distance: p2 to p1
  double td = pointLenghtFromFirst[pointBeforeTarget+1] - pointLenghtFromFirst[pointBeforeTarget];
  
  targetCoord->x = ((pd*(p2->getX()-p1->getX()))/td)+p1->getX();  

  targetCoord->y = ((pd*(p2->getY()-p1->getY()))/td)+p1->getY();

  return targetCoord;
}

void te::gm::Multi2Single(const te::gm::Geometry* g, std::vector<te::gm::Geometry*>& geoms)
{
  te::gm::GeometryCollection* gc = dynamic_cast<te::gm::GeometryCollection*>(g->clone());
  if (gc)
  {
    for (std::size_t i = 0; i < gc->getNumGeometries(); ++i)
    {
      te::gm::Geometry* currentGeom = gc->getGeometryN(i);
      currentGeom->setSRID(gc->getSRID());
      Multi2Single(currentGeom, geoms);
    }
  }
  else
    geoms.push_back(dynamic_cast<te::gm::Geometry*>(g->clone()));
}

std::vector<te::gm::Geometry*> te::gm::FixSelfIntersection(const te::gm::Geometry* g)
{
  std::vector<te::gm::Geometry*> result;

  std::unique_ptr<te::gm::Geometry> geomBuffer(g->buffer(0.0));

  std::vector<te::gm::Geometry*> vecSingleGeoms;
  te::gm::Multi2Single(geomBuffer.get(), vecSingleGeoms);

  if (g->getGeomTypeId() == te::gm::MultiPolygonType)
  {
    for (std::size_t j = 0; j < vecSingleGeoms.size(); ++j)
    {
      te::gm::MultiPolygon* pol = new te::gm::MultiPolygon(0, te::gm::MultiPolygonType, g->getSRID());

      pol->add((te::gm::Geometry*)vecSingleGeoms[j]->clone());

      result.push_back(pol);
    }
  }
  else if (g->getGeomTypeId() == te::gm::PolygonType)
  {
    for (std::size_t j = 0; j < vecSingleGeoms.size(); ++j)
    {
      result.push_back((te::gm::Geometry*)vecSingleGeoms[j]->clone());
    }
  }
  else
  {
    throw te::common::Exception(TE_TR("Unexpected geometry type!"));
  }

  return result;
}

te::gm::Geometry* te::gm::CascadedPolygonUnion(const std::vector<Polygon*>& polygonVector)
{
#ifdef TERRALIB_GEOS_ENABLED

  std::vector<geos::geom::Polygon*> geosPolygonVector;

  try
  {
    if(polygonVector.empty())
      return nullptr;

    int srid = polygonVector[0]->getSRID();

    for(auto i : polygonVector)
    {
      if(i->getGeomTypeId() == te::gm::PolygonType)
      {
        geos::geom::Polygon* geosPolygon = GEOSWriter::write(i);
        geosPolygonVector.push_back(geosPolygon);
      }
    }

    std::unique_ptr<geos::geom::Geometry> unionGeom(geos::operation::geounion::CascadedPolygonUnion::Union(&geosPolygonVector));

    te::common::FreeContents(geosPolygonVector);

    if(unionGeom == nullptr)
      return nullptr;

    unionGeom->setSRID(srid);

    return GEOSReader::read(unionGeom.get());
  }
  catch(geos::util::GEOSException& e)
  {
    te::common::FreeContents(geosPolygonVector);
    throw Exception(e.what());
  }

#else
  throw te::common::Exception(TE_TR("CascadedPolygonUnion routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Geometry* te::gm::UnaryUnion(te::gm::Geometry* geom)
{
#ifdef TERRALIB_GEOS_ENABLED

  try
  {
    std::unique_ptr<geos::geom::Geometry> geomGeos(GEOSWriter::write(geom));

    std::unique_ptr<geos::geom::Geometry> unionGeom = geos::operation::geounion::UnaryUnionOp::Union(*(geomGeos.get()));

    unionGeom->setSRID(geom->getSRID());

    return GEOSReader::read(unionGeom.get());
  }
  catch(geos::util::GEOSException& e)
  {
    throw Exception(e.what());
  }

#else
  throw te::common::Exception(TE_TR("Union routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

void te::gm::ClosestPoints(te::gm::Geometry* geomA, te::gm::Geometry* geomB,
  te::gm::Coord2D& coordA, te::gm::Coord2D& coordB)
{
#ifdef TERRALIB_GEOS_ENABLED
  std::unique_ptr<geos::geom::Geometry> thisGeomA(GEOSWriter::write(geomA));
  std::unique_ptr<geos::geom::Geometry> thisGeomB(GEOSWriter::write(geomB));

  geos::operation::distance::DistanceOp op(thisGeomA.get(), thisGeomB.get());

  geos::geom::CoordinateSequence* cs = op.closestPoints();

  if (cs->getSize() == 2)
  {
    coordA = te::gm::Coord2D(cs->getAt(0).x, cs->getAt(0).y);
    coordB = te::gm::Coord2D(cs->getAt(1).x, cs->getAt(1).y);
  }

  delete cs;

#else
  throw te::common::Exception(TE_TR("getCentroid routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Line* te::gm::GetIntersectionLine(te::gm::Geometry* geom, te::gm::Coord2D coord)
{
  std::vector<te::gm::Geometry*> geomVec;
  std::unique_ptr<te::gm::Point> pRef(new te::gm::Point(coord.x, coord.y, geom->getSRID()));
  
  te::gm::Multi2Single(geom, geomVec);

  for (std::size_t t = 0; t < geomVec.size(); ++t)
  {
    te::gm::Geometry* g = geomVec[t];

    if (g->getGeomTypeId() == te::gm::LineStringType)
    {
      te::gm::LineString* ls = static_cast<te::gm::LineString*>(g);

      if (ls->getNPoints() == 2)
      {
        std::unique_ptr<te::gm::Point> p0 = ls->getPointN(0);
        std::unique_ptr<te::gm::Point> p1 = ls->getPointN(1);

        te::gm::Line* line = new te::gm::Line(te::gm::LineStringType, geom->getSRID());

        line->setCoord(0, p0->getX(), p0->getY());
        line->setCoord(1, p1->getX(), p1->getY());

        return line;
      }
      else
      {
        for (std::size_t p = 0; p < ls->getNPoints() - 1; ++p)
        {
          std::unique_ptr<te::gm::Point> p0 = ls->getPointN(p);
          std::unique_ptr<te::gm::Point> p1 = ls->getPointN(p + 1);

          std::unique_ptr<te::gm::Line> line(new te::gm::Line(te::gm::LineStringType, geom->getSRID()));
          line->setCoord(0, p0->getX(), p0->getY());
          line->setCoord(1, p1->getX(), p1->getY());

          std::unique_ptr<te::gm::Geometry> geomMBR(te::gm::GetGeomFromEnvelope(line->getMBR(), geom->getSRID()));

          if (geomMBR->intersects(pRef.get()))
          {
            return line.release();
          }
        }
      }
    }
  }

  return 0;
}

double te::gm::GetAngle(te::gm::Coord2D coordA, te::gm::Coord2D coordB)
{
  if (coordA.x == coordB.x)
  {
    return 1.57079632679489661923;
  }

  if (coordA.y == coordB.y)
  {
    return 0.0;
  }

  double dx = (coordA.x - coordB.x);
  double dy = (coordA.y - coordB.y);

  double rad = std::atan(dy / dx);

  //Convert radians to degrees
  double angle = rad / 0.01745329251994329576;

  return angle;
}

bool te::gm::Rotate(te::gm::Coord2D pr, te::gm::LineString* l, double angle, te::gm::LineString* lOut)
{
  double alfa;
  double dx, dy;
  double x, y;
  double xr, yr;

  try
  {
    if (l->size() < 2)
      return (false);

    //transform Degree to Radius
    alfa = (4.*std::atan(1.)*angle) / 180.;

    dx = pr.getX();
    dy = pr.getY();

    for (std::size_t count = 0; count < l->size(); count++)
    {
      std::unique_ptr<te::gm::Point> curPoint = l->getPointN(count);

      x = curPoint->getX() - dx;
      y = curPoint->getY() - dy;

      xr = x * std::cos(alfa) - y * std::sin(alfa);
      yr = x * std::sin(alfa) + y * std::cos(alfa);

      lOut->setPoint(count, xr + dx, yr + dy);
    }
  }
  catch (...)
  {
    return false;
  }

  return true;
}

bool te::gm::AdjustSegment(te::gm::Point* P0, te::gm::Point* P1, double d0, te::gm::Coord2D& P0out, te::gm::Coord2D& P1out)
{
  double vL_norm1;
  double vL_norm2;

  try
  {
    if (P0->getX() == P1->getX() && P0->getY() == P1->getY())
      return false;

    te::gm::Coord2D vL1((P1->getX() - P0->getX()), (P1->getY() - P0->getY()));
    te::gm::Coord2D vL2(-1 * (P1->getX() - P0->getX()), -1 * (P1->getY() - P0->getY()));
    vL_norm1 = std::sqrt(vL1.getX() * vL1.getX() + vL1.getY() * vL1.getY());
    vL_norm2 = std::sqrt(vL2.getX() * vL2.getX() + vL2.getY() * vL2.getY());

    te::gm::Coord2D uL1((vL1.getX() / vL_norm1), (vL1.getY() / vL_norm1));
    te::gm::Coord2D uL2((vL2.getX() / vL_norm2), (vL2.getY() / vL_norm2));

    te::gm::Point* pFim1 = new te::gm::Point(P0->getX() + uL1.getX() * d0, P0->getY() + uL1.getY() * d0, P0->getSRID());
    te::gm::Point* pFim2 = new te::gm::Point(P0->getX() + uL2.getX() * d0, P0->getY() + uL2.getY() * d0, P1->getSRID());

    if (pFim1->distance(P1) <= pFim2->distance(P1))
    {
      P0out.x = P0->getX();
      P0out.y = P0->getY();
      P1out.x = pFim1->getX();
      P1out.y = pFim1->getY();
    }
    else
    {
      P0out.x = P0->getX();
      P0out.y = P0->getY();
      P1out.x = pFim2->getX();
      P1out.y = pFim2->getY();
    }
  }
  catch (...)
  {
    return false;
  }

  return true;
}

void te::gm::Polygonizer(te::gm::Geometry* g, std::vector<te::gm::Polygon*>& pols)
{
#ifdef TERRALIB_GEOS_ENABLED

  geos::operation::polygonize::Polygonizer polygonizer;

  std::unique_ptr<geos::geom::Geometry> geomGeos(GEOSWriter::write(g));
  polygonizer.add(geomGeos.get());

  std::vector <geos::geom::Polygon*>* vecpolGeos = polygonizer.getPolygons();

  for (std::size_t i = 0; i < vecpolGeos->size(); i++)
  { 
    vecpolGeos->at(i)->setSRID(g->getSRID());
    pols.push_back(GEOSReader::read(vecpolGeos->at(i)));
  }

#else
  throw te::common::Exception(TE_TR("Polygonizer routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

bool te::gm::CheckValidity(const te::gm::Geometry* geom, te::gm::TopologyValidationError& error)
{
#ifdef TERRALIB_GEOS_ENABLED

  try
  {
    std::unique_ptr<geos::geom::Geometry> g(GEOSWriter::write(geom));

    geos::operation::valid::IsValidOp vop(g.get());

    geos::operation::valid::TopologyValidationError* err = vop.getValidationError();

    if (err)
    {
      error.m_coordinate = Coord2D(err->getCoordinate().x, err->getCoordinate().y);
      error.m_message = err->getMessage();

      return false;
    }

    return true;
  }
  catch(const geos::util::GEOSException& e)
  {
    throw Exception(e.what());
  }

#else
  throw Exception(TE_TR("isValid routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Geometry* te::gm::Validate(te::gm::Geometry* geom)
{
#ifdef TERRALIB_GEOS_ENABLED

  std::unique_ptr<geos::geom::Geometry> g(GEOSWriter::write(geom));

  std::vector<te::gm::Geometry*> pAdd;
  geos::operation::polygonize::Polygonizer polygonizer;

  switch (g->getGeometryTypeId())
  {
    case geos::geom::GEOS_POLYGON:
    {
      if (g->isValid())
      {
        g->normalize(); // validate does not pick up rings in the wrong order - this will fix that
        return GEOSReader::read(g.get()); // If the polygon is valid just return it
      }

      AddPolygon(dynamic_cast<te::gm::Polygon*>(GEOSReader::read(g.get())), pAdd);
    }
    break;

    case geos::geom::GEOS_MULTIPOLYGON:
    {
      if (g->isValid()){
        g->normalize(); // validate does not pick up rings in the wrong order - this will fix that
        return GEOSReader::read(g.get()); // If the multipolygon is valid just return it
      }

      for (std::size_t n = g->getNumGeometries(); n-- > 0;)
        AddPolygon(dynamic_cast<te::gm::Polygon*>(GEOSReader::read(g->getGeometryN(n))), pAdd);

    }
    break;

    default:
      return geom;
  }

  for (std::size_t i = 0; i < pAdd.size(); i++)
    polygonizer.add(GEOSWriter::write(pAdd[i]));

  te::common::FreeContents(pAdd);

  std::vector <geos::geom::Polygon*>* vecpolGeos = polygonizer.getPolygons();

  switch (vecpolGeos->size())
  {
    case 1:
      vecpolGeos->at(0)->setSRID(geom->getSRID());
      return GEOSReader::read(vecpolGeos->at(0)); // single polygon - no need to wrap
    default:
    {
      //filter the invalid polygons
      std::size_t i = 0;
      while (i < vecpolGeos->size()) 
      {
        if (!vecpolGeos->at(i)->isValid()) 
        {
          vecpolGeos->erase(vecpolGeos->begin() + i);
        }
        else 
        {
          ++i;
        }
      }

      //polygons may still overlap! Need to sym difference them
      vecpolGeos->at(0)->setSRID(geom->getSRID());
      te::gm::Geometry* ret = GEOSReader::read(vecpolGeos->at(0));

      for (std::size_t p = 1; p < vecpolGeos->size(); p++)
      {
        vecpolGeos->at(p)->setSRID(geom->getSRID());
        ret = ret->symDifference(GEOSReader::read(vecpolGeos->at(p)));
      }

      te::common::FreeContents(*vecpolGeos);

      return ret;
    }
  }

#else
  throw Exception(TE_TR("validate routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

void te::gm::AddPolygon(te::gm::Polygon* polygon, std::vector<te::gm::Geometry*>& pAdd)
{
#ifdef TERRALIB_GEOS_ENABLED

  AddLineString(dynamic_cast<te::gm::LineString*>(polygon->getExteriorRing()), pAdd);

  for (std::size_t n = polygon->getNumInteriorRings(); n-- > 0;){
    AddLineString(dynamic_cast<te::gm::LineString*>(polygon->getInteriorRingN(n)), pAdd);
  }

#else
  throw Exception(TE_TR("addPolygon routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

void te::gm::AddLineString(te::gm::LineString* lineString, std::vector<te::gm::Geometry*>& pAdd)
{
#ifdef TERRALIB_GEOS_ENABLED

  std::unique_ptr<geos::geom::LineString> ls(GEOSWriter::write(lineString));

  // unioning the linestring with the point makes any self intersections explicit.
  geos::geom::Point* point = GEOSGeometryFactory::getGeomFactory()->createPoint(ls->getCoordinateN(0)); 

  geos::geom::Geometry* toAdd = ls->Union(point);

  pAdd.push_back(GEOSReader::read(toAdd));

#else
  throw Exception(TE_TR("addLineString routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

double te::gm::PerpendicularDistance(const te::gm::Point& first,
                                     const te::gm::Point& last,
                                     te::gm::Point& pin, te::gm::Point& pinter)
{
#ifdef TERRALIB_GEOS_ENABLED
  double d12, xmin, ymin;

  double xi = first.getX();
  double xf = last.getX();
  double yi = first.getY();
  double yf = last.getY();
  double x = pin.getX();
  double y = pin.getY();

  double dx = xf - xi;
  double dy = yf - yi;
  double a2 = (y - yi) * dx - (x - xi)*dy;

  if (dx == 0. && dy == 0.)
  {
    d12 = sqrt(((x - xi) * (x - xi)) + ((y - yi) * (y - yi)));
    d12 *= d12;
  }
  else
    d12 = a2 * a2 / (dx * dx + dy * dy);

  if (dx == 0.)
  {
    xmin = xi;
    ymin = y;
  }
  else if (dy == 0.)
  {
    xmin = x;
    ymin = yi;
  }
  else
  {
    double alfa = dy / dx;
    xmin = (x + alfa * (y - yi) + alfa * alfa * xi) / (1. + alfa * alfa);
    ymin = (x - xmin) / alfa + y;
  }

  pinter = te::gm::Point(xmin, ymin, pin.getSRID());

  return (sqrt(d12));

#else
  throw Exception(TE_TR("PerpendicularDistance routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Geometry* te::gm::PrepareGeometriesToIntersection(te::gm::Geometry* geom_A, te::gm::Geometry* geom_B)
{
#ifdef TERRALIB_GEOS_ENABLED

  te::gm::Geometry* geomResult = 0;

  switch(geom_A->getGeomTypeId())
  {
    case te::gm::PolygonType:
    {
      if(geom_B->getGeomTypeId() == te::gm::PolygonType)
      {
        te::gm::Polygon* tempPol_A = dynamic_cast<te::gm::Polygon*>(geom_A);
        te::gm::Polygon* tempPol_B = dynamic_cast<te::gm::Polygon*>(geom_B);

        geomResult = PrepareGeometriesToIntersection(*tempPol_A ,*tempPol_B);
      }
      if(geom_B->getGeomTypeId() == te::gm::MultiPolygonType)
      {
        std::unique_ptr<te::gm::MultiPolygon> mp_temp(new te::gm::MultiPolygon(0, te::gm::MultiPolygonType, geom_A->getSRID()));
        mp_temp->add(dynamic_cast<te::gm::Geometry*>(geom_A->clone()));

        geomResult = PrepareGeometriesToIntersection((te::gm::Geometry*)mp_temp.get(), geom_B);
      }

      break;
    }
    case te::gm::MultiPolygonType:
    {
      if(geom_B->getGeomTypeId() == te::gm::PolygonType)
      {
        std::unique_ptr<te::gm::MultiPolygon> mp_temp(new te::gm::MultiPolygon(0, te::gm::MultiPolygonType, geom_B->getSRID()));
        mp_temp->add(dynamic_cast<te::gm::Geometry*>(geom_B->clone()));

        geomResult = PrepareGeometriesToIntersection(geom_A, (te::gm::Geometry*)mp_temp.get());
      }
      if(geom_B->getGeomTypeId() == te::gm::MultiPolygonType)
      {
        geomResult = PrepareGeometriesToIntersection(*dynamic_cast<te::gm::MultiPolygon*>(geom_A),
                                                     *dynamic_cast<te::gm::MultiPolygon*>(geom_B));
      }

      break;
    }
    case te::gm::LineStringType:
    {
      if(geom_B->getGeomTypeId() == te::gm::LineStringType)
      {
        te::gm::LineString* tempLine_A = dynamic_cast<te::gm::LineString*>(geom_A);
        te::gm::LineString* tempLine_B = dynamic_cast<te::gm::LineString*>(geom_B);

        geomResult = PrepareGeometriesToIntersection(*tempLine_A, *tempLine_B);
      }
      if(geom_B->getGeomTypeId() == te::gm::MultiLineStringType)
      {
        std::unique_ptr<te::gm::LineString> tempLine_A(dynamic_cast<te::gm::LineString*>(geom_A->clone()));

        te::gm::MultiLineString* geometryCollection_B = dynamic_cast<te::gm::MultiLineString*>(geom_B);
        std::vector<te::gm::Geometry*>& geometryVector = geometryCollection_B->getGeometries();

        for(std::size_t i = 0; i < geometryVector.size(); ++i)
        {
          tempLine_A.reset(PrepareGeometriesToIntersection(
              *tempLine_A.get(),
              *dynamic_cast<te::gm::LineString*>(geometryVector[i])));
        }

        geomResult = tempLine_A.release();
      }

      break;
    }
    default:
      break;
  }

  return geomResult;

#else
  throw Exception(TE_TR("PrepareGeometriesToIntersection routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Geometry* te::gm::PrepareGeometriesToIntersection(const te::gm::MultiPolygon& mpol_A,
                                                         const te::gm::MultiPolygon& mpol_B)
{
#ifdef TERRALIB_GEOS_ENABLED

  te::gm::Geometry* geomResult = 0;

  const std::vector<te::gm::Geometry*>& geometryVectorA = mpol_A.getGeometries();
  const std::vector<te::gm::Geometry*>& geometryVectorB = mpol_B.getGeometries();

  std::unique_ptr<te::gm::MultiPolygon> geometryCollection_temp(
      new te::gm::MultiPolygon(0, te::gm::MultiPolygonType, mpol_A.getSRID()));

  std::vector<te::gm::Geometry*> geometryVectorACloned;
  for (std::size_t i = 0; i < geometryVectorA.size(); ++i)
  {
    geometryVectorACloned.push_back((te::gm::Geometry*)geometryVectorA[i]->clone());
  }

  for(std::size_t i = 0; i < geometryVectorACloned.size(); ++i)
  {
    std::unique_ptr<te::gm::Geometry> geometryA((te::gm::Geometry*)geometryVectorACloned[i]->clone());

    for(std::size_t j = 0; j < geometryVectorB.size(); ++j)
    {
      geometryA.reset(PrepareGeometriesToIntersection(
          *dynamic_cast<te::gm::Polygon*>(geometryA.get()),
          *dynamic_cast<te::gm::Polygon*>(geometryVectorB[j])));

      if (geometryA->getGeomTypeId() == te::gm::MultiPolygonType)
      {
        std::unique_ptr<te::gm::MultiPolygon> resultMultiA(dynamic_cast<te::gm::MultiPolygon*>(geometryA->clone()));

        for (std::size_t k = 0; k < resultMultiA->getNumGeometries(); ++k)
        {
          te::gm::Geometry* geomResultA = resultMultiA->getGeometryN(k);
          if (k == 0)
          {
            geometryA.reset((te::gm::Geometry*)geomResultA->clone());
          }
          else
          {
            geometryVectorACloned.push_back((te::gm::Geometry*)geomResultA->clone());
          }
        }
      }
    }

    geometryCollection_temp->add(geometryA.release());
  }

  geomResult = geometryCollection_temp.release();

  te::common::FreeContents(geometryVectorACloned);

  return geomResult;

#else
  throw Exception(TE_TR("PrepareGeometriesToIntersection routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Geometry* te::gm::PrepareGeometriesToIntersection(const te::gm::Polygon& pol_A,
                                                         const te::gm::Polygon& pol_B)
{
#ifdef TERRALIB_GEOS_ENABLED

  const te::gm::Envelope* env_A = pol_A.getMBR();
  te::gm::Envelope expandedEnv_A;
  expandedEnv_A.init(env_A->getLowerLeftX() - tolerance,
                     env_A->getLowerLeftY() - tolerance,
                     env_A->getUpperRightX() + tolerance,
                     env_A->getUpperRightY() + tolerance);

  const te::gm::Envelope* env_B = pol_B.getMBR();
  te::gm::Envelope expandedEnv_B;
  expandedEnv_B.init(env_B->getLowerLeftX() - tolerance,
                     env_B->getLowerLeftY() - tolerance,
                     env_B->getUpperRightX() + tolerance,
                     env_B->getUpperRightY() + tolerance);

  if(!expandedEnv_A.intersects(expandedEnv_B))
    return dynamic_cast<te::gm::Polygon*>(pol_A.clone());

  std::unique_ptr<te::gm::Polygon> pol_Result(new te::gm::Polygon(0, te::gm::PolygonType, pol_A.getSRID()));

  const std::vector<te::gm::Curve*>& vecCurve_A = pol_A.getRings();
  const std::vector<te::gm::Curve*>& vecCurve_B = pol_B.getRings();

  for (std::size_t i = 0; i < vecCurve_A.size(); ++i)
  {
    te::gm::LineString* ls_A = dynamic_cast<te::gm::LineString*>(vecCurve_A[i]);
    std::unique_ptr<te::gm::LineString> ls_Result(dynamic_cast<te::gm::LineString*>(ls_A->clone()));

    for (std::size_t j = 0; j < vecCurve_B.size(); ++j)
    {
      te::gm::LineString* ls_B = dynamic_cast<te::gm::LineString*>(vecCurve_B[j]);

      ls_Result.reset(PrepareGeometriesToIntersection(*ls_Result.get(), *ls_B));
    }

    pol_Result->add(ls_Result.release());
  }

  if (pol_Result->isValid() == true)
  {
    return pol_Result.release();
  }

  te::gm::Geometry* result = pol_Result->buffer(0);

  return result;

#else
  throw Exception(TE_TR("PrepareGeometriesToIntersection routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::LineString* te::gm::PrepareGeometriesToIntersection(const te::gm::LineString& ls_A,
                                                            const te::gm::LineString& ls_B)
{
#ifdef TERRALIB_GEOS_ENABLED

  // Check if MBRs of both LineStrings have intersection.
  const te::gm::Envelope* env_A = ls_A.getMBR();
  te::gm::Envelope expandedEnv_A;
  expandedEnv_A.init(env_A->getLowerLeftX() - tolerance,
                     env_A->getLowerLeftY() - tolerance,
                     env_A->getUpperRightX() + tolerance,
                     env_A->getUpperRightY() + tolerance);

  const te::gm::Envelope* env_B = ls_B.getMBR();
  te::gm::Envelope expandedEnv_B;
  expandedEnv_B.init(env_B->getLowerLeftX() - tolerance,
                     env_B->getLowerLeftY() - tolerance,
                     env_B->getUpperRightX() + tolerance,
                     env_B->getUpperRightY() + tolerance);

  if(!expandedEnv_A.intersects(expandedEnv_B))
    return dynamic_cast<te::gm::LineString*>(ls_A.clone());

  // Creating the RTree with the ls_B segment boxes intersects ls_A segment box
  te::sam::rtree::Index<std::size_t, 8> rtree;

  std::size_t npoints_B = ls_B.getNPoints();
  std::size_t n_segments_B = npoints_B - 1;
  std::vector<te::gm::LineString> segments_B;

  for(std::size_t i = 0; i < n_segments_B; ++i)
  {
    te::gm::LineString segment_B(2, te::gm::LineStringType, ls_A.getSRID());

    std::unique_ptr<Point> pStart = ls_B.getPointN(i);
    segment_B.setPointN(0, *pStart.get());

    std::unique_ptr<Point> pEnd = ls_B.getPointN(i+1);
    segment_B.setPointN(1, *pEnd.get());


    const te::gm::Envelope* segmentEnv_B = segment_B.getMBR();
    te::gm::Envelope expandedSegmentEnv_B;
    expandedSegmentEnv_B.init(segmentEnv_B->getLowerLeftX() - tolerance,
                              segmentEnv_B->getLowerLeftY() - tolerance,
                              segmentEnv_B->getUpperRightX() + tolerance,
                              segmentEnv_B->getUpperRightY() + tolerance);

    if(expandedEnv_A.intersects(expandedSegmentEnv_B))
    {
      rtree.insert(*segmentEnv_B, segments_B.size());
      segments_B.push_back(segment_B);
    }
  }

  // First try to snap the points from ls_A in the points of ls_B
  std::unique_ptr<te::gm::LinearRing> lr_Result(dynamic_cast<te::gm::LinearRing*>(ls_A.clone()));
  lr_Result.reset(SnapLineToPoints(rtree, segments_B, *lr_Result.get()));

  lr_Result.reset(AddIntersectionPoints(rtree, segments_B, *lr_Result.get(), true));
  lr_Result.reset(AddIntersectionPoints(rtree, segments_B, *lr_Result.get()));

  return (te::gm::LineString*)lr_Result.release();

#else
  throw Exception(TE_TR("PrepareGeometriesToIntersection routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::LinearRing* te::gm::SnapLineToPoints(const te::sam::rtree::Index<std::size_t, 8>& rtree,
                                             const std::vector<te::gm::LineString>& referenceSegments,
                                             const te::gm::LinearRing& linearRingToSnap)
{
#ifdef TERRALIB_GEOS_ENABLED

  te::gm::LinearRing* lr_Result = dynamic_cast<te::gm::LinearRing*>(linearRingToSnap.clone());

  std::size_t npoints_A = lr_Result->getNPoints();

  for (std::size_t i = 0; i < npoints_A; ++i)
  {
    std::unique_ptr<Point> currentPoint = lr_Result->getPointN(i);

    te::gm::Envelope expandedEnv;
    expandedEnv.init(
        currentPoint->getX() - tolerance, currentPoint->getY() - tolerance,
        currentPoint->getX() + tolerance, currentPoint->getY() + tolerance);

    std::vector<size_t> report;
    rtree.search(expandedEnv, report);

    if (report.empty())
      continue;

    // Extracts all the points.
    std::vector<te::gm::Point> candidates;
    for (std::size_t ls = 0; ls < report.size(); ++ls)
    {
      std::size_t index = report[ls];

      std::unique_ptr<Point> pStart = referenceSegments[index].getPointN(0);
      candidates.push_back(*pStart.get());

      std::unique_ptr<Point> pEnd = referenceSegments[index].getPointN(1);
      candidates.push_back(*pEnd.get());
    }

    //Checks if there is on point that is within the tolerance.
    //If there is more than one, is found the closest.
    DistanceOrderFunctor func(*currentPoint);
    std::sort(candidates.begin(), candidates.end(), func);

    double distance = currentPoint->distance(&candidates[0]);
    if(distance != 0 && distance <= tolerance)
    {
      lr_Result->setPointN(i, candidates[0]);
      if (i == 0)
      {
        lr_Result->setPointN(npoints_A - 1, candidates[0]);
      }
    }
  }

  return lr_Result;

#else
  throw Exception(TE_TR("SnapLineToPoints routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::LinearRing* te::gm::AddIntersectionPoints(const te::sam::rtree::Index<std::size_t, 8>& rtree,
                                                  const std::vector<te::gm::LineString>& candidateSegments,
                                                  const te::gm::LinearRing& lr_Reference,
                                                  const bool& usePerpendicularDistance)
{
#ifdef TERRALIB_GEOS_ENABLED

  std::unique_ptr<te::gm::LinearRing> lr_Ref(dynamic_cast<te::gm::LinearRing*>(lr_Reference.clone()));

  std::size_t n_segments_A = lr_Ref->getNPoints() - 1;

  std::vector<te::gm::Point> vecOutputPoints;

  for (std::size_t i = 0; i < n_segments_A; ++i)
  {
    // Search for candidates.
    te::gm::LineString segment_A(2, te::gm::LineStringType, lr_Ref->getSRID());

    std::unique_ptr<Point> pStart = lr_Ref->getPointN(i);
    segment_A.setPointN(0, *pStart.get());

    std::unique_ptr<Point> pEnd = lr_Ref->getPointN(i + 1);
    segment_A.setPointN(1, *pEnd.get());

    vecOutputPoints.push_back(*pStart.get());

    const te::gm::Envelope* env = segment_A.getMBR();

    std::unique_ptr<te::gm::Envelope> expandedEnv(new te::gm::Envelope());
    expandedEnv->init(env->getLowerLeftX() - tolerance,
                     env->getLowerLeftY() - tolerance,
                     env->getUpperRightX() + tolerance,
                     env->getUpperRightY() + tolerance);

    std::vector<size_t> report;
    rtree.search(*expandedEnv.get(), report);

    if (report.empty())
      continue;

    std::vector<te::gm::LineString> candidates;
    for(std::size_t ls = 0; ls < report.size(); ++ls)
    {
      std::size_t index = report[ls];
      candidates.push_back(candidateSegments[index]);
    }

    // Executes the intersection for each candidate point, and get the coordinates.
    std::vector<te::gm::Point> vecIntersectionPoints;

    if(usePerpendicularDistance)
      vecIntersectionPoints = GetIntersectionPointsByPerpendicularDistance(segment_A, candidates);
    else
      vecIntersectionPoints = GetIntersectionPointsByOperators(segment_A, candidates);

    DistanceOrderFunctor func(*pStart.get());
    std::sort(vecIntersectionPoints.begin(), vecIntersectionPoints.end(), func);

    for (std::size_t j = 0; j < vecIntersectionPoints.size(); ++j)
    {
      te::gm::Point lastPoint = vecOutputPoints[vecOutputPoints.size() - 1];

      if (!vecIntersectionPoints[j].equals(&lastPoint) &&
          !vecIntersectionPoints[j].equals(pStart.get()) &&
          !vecIntersectionPoints[j].equals(pEnd.get()))
        vecOutputPoints.push_back(vecIntersectionPoints[j]);
    }
  }

  // Set the end point in the output vector.
  std::unique_ptr<Point> pStart = lr_Ref->getStartPoint();
  vecOutputPoints.push_back(*pStart.get());

  te::gm::LinearRing* result = new te::gm::LinearRing(vecOutputPoints.size(), te::gm::LineStringType, lr_Ref->getSRID());

  for (std::size_t i = 0; i < vecOutputPoints.size(); ++i)
    result->setPointN(i, vecOutputPoints[i]);

  return result;

#else
  throw Exception(TE_TR("GetIntersectionPoints routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

std::vector<te::gm::Point> te::gm::GetIntersectionPointsByOperators(
    const te::gm::LineString& lsReference,
    const std::vector<te::gm::LineString>& candidates)
{
#ifdef TERRALIB_GEOS_ENABLED

  std::vector<te::gm::Point> intersectionPoints;

  std::unique_ptr<te::gm::Point> startPoint = lsReference.getStartPoint();
  std::unique_ptr<te::gm::Point> endPoint = lsReference.getEndPoint();

  for (std::size_t i = 0; i < candidates.size(); ++i)
  {
    std::unique_ptr<te::gm::Point>candidateStartPoint = candidates[i].getStartPoint();
    std::unique_ptr<te::gm::Point>candidateEndPoint = candidates[i].getEndPoint();

    std::unique_ptr<te::gm::Geometry> intersection(
        lsReference.intersection(&candidates[i]));

    if(intersection == nullptr)
      continue;

    switch (intersection->getGeomTypeId())
    {
      case te::gm::PointType:
      {
        te::gm::Point p(*dynamic_cast<te::gm::Point*>(intersection.get()));

        if (p.equals(startPoint.get()) == true || p.equals(endPoint.get()) == true)
        {
          break;
        }

        // If the calculated intersection point is "near" the start or end point of the current segment, just snap it.

        double distanceStartPointB = p.distance(candidateStartPoint.get());
        double distanceEndPointB = p.distance(candidateEndPoint.get());

        if (distanceStartPointB <= tolerance)
        {
          intersectionPoints.push_back(*candidateStartPoint.get());
        }
        else if (distanceEndPointB <= tolerance)
        {
          intersectionPoints.push_back(*candidateEndPoint.get());
        }
        else
        {
          intersectionPoints.push_back(p);
        }

        break;
      }
      case te::gm::LineStringType:
      {
        te::gm::LineString* l = dynamic_cast<te::gm::LineString*>(intersection.get());

        for (std::size_t n = 0; n < l->getNPoints(); ++n)
        {
          std::unique_ptr<te::gm::Point> p = l->getPointN(n);

          if (p->equals(startPoint.get()) == true || p->equals(endPoint.get()) == true)
          {
            continue;
          }

          // If the calculated intersection point is "near"  the start or end point of the current segment, just snap it.
          double distanceStartPointB = p->distance(candidateStartPoint.get());
          double distanceEndPointB = p->distance(candidateEndPoint.get());

          if (distanceStartPointB <= tolerance)
          {
            intersectionPoints.push_back(*candidateStartPoint.get());
          }
          else if (distanceEndPointB <= tolerance)
          {
            intersectionPoints.push_back(*candidateEndPoint.get());
          }
          else
          {
            intersectionPoints.push_back(*p.get());
          }
        }
        break;
      }
      default:
        break;
    }
  }

  return intersectionPoints;

#else
  throw Exception(TE_TR("GetIntersectionPointsByOperators routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

std::vector<te::gm::Point> te::gm::GetIntersectionPointsByPerpendicularDistance(
    const te::gm::LineString& lsReference,
    const std::vector<te::gm::LineString>& candidates)
{
#ifdef TERRALIB_GEOS_ENABLED

  std::vector<te::gm::Point> intersectionPoints;

  std::unique_ptr<te::gm::Point> startPoint = lsReference.getStartPoint();
  std::unique_ptr<te::gm::Point> endPoint = lsReference.getEndPoint();

  for (std::size_t i = 0; i < candidates.size(); ++i)
  {
    std::unique_ptr<te::gm::Point>candidateStartPoint = candidates[i].getStartPoint();
    std::unique_ptr<te::gm::Point>candidateEndPoint = candidates[i].getEndPoint();

    std::unique_ptr<geos::geom::LineSegment> lineSegment(
        new geos::geom::LineSegment(startPoint->getX(), startPoint->getY(),
                                    endPoint->getX(), endPoint->getY()));


    te::gm::Point pinter;

    double startDistance = te::gm::PerpendicularDistance(
        *startPoint.get(), *endPoint.get(), *candidateStartPoint.get(), pinter);

    std::unique_ptr<geos::geom::Coordinate> coordStart(
        new geos::geom::Coordinate(pinter.getX(), pinter.getY()));

    double projFactorStart = lineSegment->projectionFactor(*coordStart.get());

    if(startDistance <= tolerance && (projFactorStart >= 0 && projFactorStart <= 1))
      intersectionPoints.push_back(*candidateStartPoint.get());



    double endDistance = te::gm::PerpendicularDistance(
        *startPoint.get(), *endPoint.get(), *candidateEndPoint.get(), pinter);

    std::unique_ptr<geos::geom::Coordinate> coordEnd(
        new geos::geom::Coordinate(pinter.getX(), pinter.getY()));

    double projFactorEnd = lineSegment->projectionFactor(*coordEnd.get());

    if(endDistance <= tolerance && (projFactorEnd >= 0 && projFactorEnd <= 1))
      intersectionPoints.push_back(*candidateEndPoint.get());
  }

  return intersectionPoints;

#else
  throw Exception(TE_TR("GetIntersectionPointsByPerpendicularDistance routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::LineString te::gm::CreateLine(const te::gm::Coord2D& startCoord, const te::gm::Coord2D& endCoord, const int& srid, const int& numberOfIntermediateCoords)
{
  double gapX = (endCoord.getX() - startCoord.getX()) / (double)(numberOfIntermediateCoords + 1);
  double gapY = (endCoord.getY() - startCoord.getY()) / (double)(numberOfIntermediateCoords + 1);

  std::vector<te::gm::Coord2D> coords;

  coords.push_back(startCoord);
  for (std::size_t i = 0; i < numberOfIntermediateCoords; ++i)
  {
    std::size_t coordIndex = i + 1;
    te::gm::Coord2D coord(startCoord.getX() + (coordIndex * gapX), startCoord.getY() + (coordIndex * gapY));
    coords.push_back(coord);
  }
  coords.push_back(endCoord);

  te::gm::LineString resultLine(coords.size(), te::gm::LineStringType, srid);

  for (std::size_t t = 0; t < coords.size(); ++t)
    resultLine.setPoint(t, coords[t].getX(), coords[t].getY());

  return resultLine;
}

te::gm::Polygon te::gm::CreatePolygon(const te::gm::Envelope& box, const int& srid, const int& numberOfIntermediateCoords)
{
  te::gm::Coord2D lowerLeft(box.m_llx, box.m_lly);
  te::gm::Coord2D lowerRight(box.m_urx, box.m_lly);

  te::gm::Coord2D upperLeft(box.m_llx, box.m_ury);
  te::gm::Coord2D upperRight(box.m_urx, box.m_ury);

  te::gm::LineString bottomLine = CreateLine(lowerLeft, lowerRight, srid, numberOfIntermediateCoords);
  te::gm::LineString leftLine = CreateLine(lowerRight, upperRight, srid, numberOfIntermediateCoords);
  te::gm::LineString topLine = CreateLine(upperRight, upperLeft, srid, numberOfIntermediateCoords);
  te::gm::LineString rightLine = CreateLine(upperLeft, lowerLeft, srid, numberOfIntermediateCoords);

  std::vector<te::gm::Coord2D> coords;

  //we do not add the last coord to avoid duplicity
  for (std::size_t i = 0; i < bottomLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(bottomLine.getX(i), bottomLine.getY(i)));

  for (std::size_t i = 0; i < leftLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(leftLine.getX(i), leftLine.getY(i)));

  for (std::size_t i = 0; i < topLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(topLine.getX(i), topLine.getY(i)));

  for (std::size_t i = 0; i < rightLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(rightLine.getX(i), rightLine.getY(i)));

  te::gm::LineString* fullLine = new te::gm::LineString(coords.size(), te::gm::LineStringType);

  for (std::size_t t = 0; t < coords.size(); ++t)
    fullLine->setPoint(t, coords[t].getX(), coords[t].getY());

  te::gm::Polygon polygon(1, te::gm::PolygonType, srid);
  polygon.setRingN(0, fullLine);

  return polygon;
}
